/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/

#include "conference.h"

void con_room_log(cnr room, char *nick, char *message)
{
    time_t t;
    pool p;
    xmlnode xml;
    jid user;
    char *tstr;
    FILE *logfile = room->logfile;

    if(logfile == NULL || message == NULL) 
	return;

    p = pool_new();

    /* nicked from mod_time */
    t = time(NULL);
    tstr = ctime(&t);
    tstr[j_strlen(tstr) - 1] = '\0'; /* cut off newline */

    if(room->logformat == LOG_XML)
    {
	user = jid_new(p, jid_full(room->id));
	jid_set(user, nick, JID_RESOURCE);

	xml = jutil_msgnew("groupchat", jid_full(room->id) , NULL, message);
	xmlnode_put_attrib(xml, "from", jid_full(user));
	jutil_delay(xml, NULL);

	fprintf(logfile, "%s\n", xmlnode2str(xml));

        xmlnode_free(xml);
    }
    else if(room->logformat == LOG_XHTML)
    {
        if(nick)
        {
	    fprintf(logfile, "%s &gt;%s&lt; %s<br\\>\n", tstr, nick, message);
        }
        else
        {
	    fprintf(logfile, "%s --- %s<br\\>\n", tstr, message);
        }
    }
    else
    {
        if(nick)
        {
	    fprintf(logfile, "%s <%s> %s\n", tstr, nick, message);
        }
        else
        {
	    fprintf(logfile, "%s --- %s\n", tstr, message);
        }
    }

    fflush(logfile);
    pool_free(p);
    return;
}

void con_room_log_new(cnr room)
{
    char *filename;
    char *logdir;
    char *ext;
    struct stat fileinfo;

    int type = room->logformat;
    char *dirname = jid_full(room->id);
    spool sp = spool_new(room->p);

    if(room->master->logdir)
    {
	logdir = room->master->logdir;
    }
    else
    {
	logdir = j_strdup(".");
    }

    spooler(sp, logdir, "/", dirname, sp);
    filename = spool_print(sp);

    if(stat(filename,&fileinfo) < 0 && mkdir(filename, S_IRWXU) < 0)
    {
	log_error(NAME, "[%s] ERR: unable to open log directory >%s<", FZONE, filename);
	return;
    }

    if(type == LOG_XML)
	ext = j_strdup("xml");
    else if(type == LOG_XHTML)
	ext = j_strdup("html");
    else
	ext = j_strdup("txt");

    spooler(sp, "/", dateget(), ".", ext, sp);

    filename = spool_print(sp);

    if(stat(filename,&fileinfo) < 0)
    {
	log_debug(NAME, "[%s] New logfile >%s<", FZONE, filename);

        room->logfile = fopen(filename, "a");

	if(type == LOG_XHTML)
	{
	    fprintf(room->logfile, "<html>\n<head>\n<title>Logs for %s, %s<\\title>\n<\\head>\n<body>\n", jid_full(room->id), dateget());
            fflush(room->logfile);
	}
    }
    else
    {
        room->logfile = fopen(filename, "a");
    }


    if(room->logfile == NULL)
	log_error(NAME, "[%s] ERR: unable to open log file >%s<", FZONE, filename);
    else
	log_debug(NAME, "[%s] Opened logfile >%s<", FZONE, filename);


    return;
}

void con_room_log_close(cnr room)
{
    int type = room->logformat;
    FILE *logfile = room->logfile;

    if(logfile == NULL)
        return;

    log_debug(NAME, "[%s] Closing logfile for room >%s<", FZONE, jid_full(room->id));

    if(type == LOG_XHTML)
    {
        fprintf(logfile, "<\\body>\n<\\html>\n");
        fflush(logfile);
    }

    fclose(room->logfile);
}

void con_room_send_invite(cnu sender, xmlnode node)
{
    xmlnode result;
    xmlnode element;
    xmlnode invite;
    char *body, *user, *reason, *inviter;
    cnr room;
    jid from;
    pool p;

    if(sender == NULL || node == NULL)
	    return;

    log_debug(NAME, "[%s] Sending room invite", FZONE);

    room = sender->room;
    from = sender->realid;

    invite = xmlnode_get_tag(node, "invite");
    user = xmlnode_get_attrib(invite, "to");
    reason = xmlnode_get_tag_data(invite, "reason");

    if(room->public == 1)
    {
        inviter = jid_full(jid_user(from));
    }
    else
    {
	inviter = xmlnode_get_data(sender->nick);
    }

    xmlnode_put_attrib(invite, "from", inviter);
    xmlnode_hide_attrib(invite, "to");

    p = xmlnode_pool(node);

    if(reason == NULL)
    {
	reason = spools(p, "None given", p);
    }

    body = spools(p, "You have been invited to the ", jid_full(room->id), " room by ", inviter, "\nReason: ", reason, p);

    result = jutil_msgnew("normal", user , "Invitation", body);
    xmlnode_put_attrib(result, "from", jid_full(room->id));


    if(room->secret != NULL)
    {
	xmlnode_insert_cdata(xmlnode_insert_tag(invite, "password"), room->secret, -1);
    }

    xmlnode_insert_node(result, node);

    element = xmlnode_insert_tag(result, "x");
    xmlnode_put_attrib(element, "jid", jid_full(room->id));
    xmlnode_put_attrib(element, "xmlns", NS_X_CONFERENCE);
    xmlnode_insert_cdata(element, reason, -1);

    log_debug(NAME, "[%s] >>>%s<<<", FZONE, xmlnode2str(result));

    deliver(dpacket_new(result), NULL);
    return;
    
}    

void con_room_leaveall(xht h, const char *key, void *data, void *arg)
{
    cnu user = (cnu)data;
    xmlnode info = (xmlnode)arg;
    char *alt, *reason;
    xmlnode presence;
    xmlnode tag;
    xmlnode element;
    xmlnode node;

    presence = jutil_presnew(JPACKET__UNAVAILABLE, NULL, NULL);
    tag = xmlnode_insert_tag(presence,"x");
    xmlnode_put_attrib(tag, "xmlns", NS_MUC_USER);

    element = xmlnode_insert_tag(tag, "item");
    xmlnode_put_attrib(element, "role", "none");
    xmlnode_put_attrib(element, "affiliation", "none");

    if (info != NULL)
    {
	reason = xmlnode_get_tag_data(info, "reason");
	node = xmlnode_insert_tag(element, "reason");

	if(reason != NULL)
	{
            xmlnode_insert_cdata(node, reason, -1);
	}

        alt = xmlnode_get_attrib(xmlnode_get_tag(info, "alt"), "jid");

	if (alt != NULL)
	{
	    xmlnode_put_attrib(xmlnode_insert_tag(tag, "alt"), "jid", alt);
        }
    }

    element = xmlnode_insert_tag(tag, "status");
    xmlnode_put_attrib(element, "code", STATUS_MUC_DESTROY);

    con_user_send(user, user, presence);
}

void _con_room_usernick(xht h, const char *key, void *data, void *arg)
{
    cnu u = (cnu)data;
    xmlnode x = (xmlnode)arg;
    if(j_strcmp(xmlnode_get_data(x),xmlnode_get_data(u->nick)) == 0)
        xmlnode_put_vattrib(x, "u", (void*)u);
}

cnu con_room_usernick(cnr room, char *nick)
{
    cnu user;
    xmlnode node = xmlnode_new_tag("nick");

    log_debug(NAME, "[%s] searching for nick %s in room %s", FZONE, nick, jid_full(room->id));

    if(room == NULL)
    {
	return NULL;
    }

    xmlnode_insert_cdata(node, nick, -1);
    xhash_walk(room->local, _con_room_usernick, (void *)node);

    user = (cnu)xmlnode_get_vattrib(node,"u");
    xmlnode_free(node);
    return user;
}

/* returns a valid nick from the list, x is the first <nick>foo</nick> xmlnode, checks siblings */
char *con_room_nick(cnr room, cnu user, xmlnode x)
{
    char *nick = NULL;
    xmlnode cur;
    int count = 1;

    log_debug(NAME, "[%s] looking for valid nick in room %s from starter %s", FZONE, jid_full(room->id), xmlnode2str(x));

    /* make-up-a-nick phase */
    if(x == NULL)
    {
        nick = pmalloco(user->p, j_strlen(user->realid->user) + 10);
        sprintf(nick,"%s",user->realid->user);
        while(con_room_usernick(room, nick) != NULL)
            sprintf(nick,"%s%d",user->realid->user,count++);
        return nick;
    }

    /* scan the list */
    for(cur = x; cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(j_strcmp(xmlnode_get_name(cur),"nick") == 0 && (nick = xmlnode_get_data(cur)) != NULL)
            if(con_room_usernick(room, nick) == NULL)
                break;
    }

    if(is_registered(room->master, jid_full(jid_user(user->realid)), nick) == -1)
	nick = NULL;

    return nick;
}

void con_room_browsewalk(xht h, const char *key, void *data, void *arg)
{
    cnu u = (cnu)data;
    xmlnode q = (xmlnode)arg;
    xmlnode x = xmlnode_insert_tag(q,"user");
    xmlnode_put_attrib(x,"name",xmlnode_get_data(u->nick));
    xmlnode_put_attrib(x,"jid",jid_full(u->localid));
}

void con_room_sendwalk(xht h, const char *key, void *data, void *arg)
{
    xmlnode x = (xmlnode)arg;
    cnu to = (cnu)data;
    cnu from = (cnu)xmlnode_get_vattrib(x,"cnu");
    xmlnode output;

    if(j_strncmp(xmlnode_get_name(x),"presence",8) == 0)
	    output = add_extended_presence(from, to, x, NULL, NULL, NULL);
    else
	    output = x;

    con_user_send(to, from, xmlnode_dup(output)); /* Need to send duplicate */
}


void con_room_outsider(cnr room, cnu from, jpacket jp)
{
    xmlnode q;
    xmlnode result;

    log_debug(NAME, "[%s] handling request from outsider %s to room %s", FZONE, jid_full(jp->from), jid_full(room->id));

    /* any presence here is just fluff */
    if(jp->type == JPACKET_PRESENCE)
    {
	log_debug(NAME, "[%s] Dropping presence from outsider", FZONE);
        xmlnode_free(jp->x);
        return;
    }

    if(jp->type == JPACKET_MESSAGE)
    {
	log_debug(NAME, "[%s] Bouncing message from outsider", FZONE);
        jutil_error(jp->x, TERROR_FORBIDDEN);
        deliver(dpacket_new(jp->x), NULL);
        return;
    }

    /* public iq requests */
    if(jpacket_subtype(jp) == JPACKET__SET)
    {
	if(NSCHECK(jp->iq, NS_MUC_OWNER))
	{
	    log_debug(NAME, "[%s] IQ Set for owner function", FZONE);

	    if(from && is_owner(room, jp->from))
	    {       
	        xdata_room_config(room, from, room->locked, jp->x);

		jutil_iqresult(jp->x);
		deliver(dpacket_new(jp->x), NULL);

	    }   
	    else
	    {
                jutil_error(jp->x, TERROR_NOTALLOWED);
                deliver(dpacket_new(jp->x), NULL);
	    }

	    return;
	}
    }

    if(jpacket_subtype(jp) == JPACKET__GET)
    {
        if(NSCHECK(jp->iq, NS_BROWSE))
        {
            jutil_iqresult(jp->x);
            q = xmlnode_insert_tag(jp->x,"item");
            xmlnode_put_attrib(q,"category","conference");

            if(room->public && room->invitation == 0)
            {
                xmlnode_put_attrib(q,"type","public");
                xhash_walk(room->local, con_room_browsewalk, (void*)q);
            }
	    else if(room->public && is_member(room, jp->from))
            {
                xmlnode_put_attrib(q,"type","public");
                xhash_walk(room->local, con_room_browsewalk, (void*)q);
            }
	    else
	    {
                xmlnode_put_attrib(q,"type","private");
            }

            xmlnode_put_attrib(q,"xmlns",NS_BROWSE);
            xmlnode_put_attrib(q,"name",room->name);
            xmlnode_put_attrib(q,"version",VERSION);

            xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_MUC,-1);

            deliver(dpacket_new(jp->x), NULL);
            return;
	}
	else if(NSCHECK(jp->iq, NS_DISCO_INFO))
	{
	    log_debug(NAME, "[%s] Outside room packet - Disco Request", FZONE);

	    jutil_iqresult(jp->x);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"), "xmlns", NS_DISCO_INFO);
	    jpacket_reset(jp);

	    result = xmlnode_insert_tag(jp->iq,"identity");
	    xmlnode_put_attrib(result, "category", "conference");
	    xmlnode_put_attrib(result, "type", "text");
	    xmlnode_put_attrib(result, "name", room->name);

	    xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "type", NS_GROUPCHAT);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "type", NS_MUC);

	    deliver(dpacket_new(jp->x), NULL);
	    return;
	}
	else if(NSCHECK(jp->iq, NS_MUC_OWNER))
        {
            if(j_strcmp(xmlnode_get_name(jp->iq),"query") == 0)
            {
	        log_debug(NAME, "[%s] IQ Get for owner: configuration", FZONE);

	        if(!is_owner(room, from->realid))
	        {
		    jutil_error(jp->x, TERROR_BAD);
		    deliver(dpacket_new(jp->x), NULL);
	            return;
	        }

	        xdata_room_config(room, from, 0, jp->x);
		return;
	    }
        }
    }

    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x), NULL);

}

void con_room_process(cnr room, cnu from, jpacket jp)
{
    char *nick;
    char *key;
    xmlnode result, item, x, node;
    jid id;

    log_debug(NAME, "[%s] handling request from participant %s(%s/%s) to room %s", FZONE, jid_full(from->realid),from->localid->resource,xmlnode_get_data(from->nick),jid_full(room->id));

    /* presence is just stored and forwarded */
    if(jp->type == JPACKET_PRESENCE)
    {
        xmlnode_free(from->presence);
	from->presence = xmlnode_dup(jp->x);
	jutil_delay(from->presence, NULL);

        xmlnode_put_vattrib(jp->x,"cnu",(void*)from);
        xhash_walk(room->local,con_room_sendwalk,(void*)jp->x);

	xmlnode_free(jp->x);
        return;
    }

    if(jp->type == JPACKET_MESSAGE)
    {
	if(NSCHECK(xmlnode_get_tag(jp->x,"x"),NS_MUC_USER))
	{
	    /* Invite */
	    log_debug(NAME, "[%s] Found invite request", FZONE);

	    if(room->invitation == 1 && room->invites == 0 && !is_admin(room, from->realid))
	    {
	        log_debug(NAME, "[%s] Forbidden invitation request, returning error", FZONE);

	        jutil_error(jp->x,TERROR_FORBIDDEN);
		deliver(dpacket_new(jp->x),NULL);
		return;
	    }

	    item = xmlnode_dup(xmlnode_get_tag(jp->x,"x"));
	    nick = xmlnode_get_attrib(xmlnode_get_tag(item, "invite"), "to");

	    if( nick == NULL)
	    {
	        log_debug(NAME, "[%s] No receipient, returning error", FZONE);

	        jutil_error(jp->x,TERROR_BAD);
		deliver(dpacket_new(jp->x),NULL);
		return;
	    }

	    if(room->invitation == 1)
	    {
	        id = jid_new(xmlnode_pool(item), nick);
	        key = j_strdup(jid_full(jid_user(id)));
	        xhash_put(room->member, key, (void*)item);
	    }

	    con_room_send_invite(from, xmlnode_get_tag(jp->x,"x"));

	    xmlnode_free(jp->x);
	    return;
	}

        /* if topic, store it */
        if((x = xmlnode_get_tag(jp->x,"subject")) != NULL )
        {

            /* catch the invite hack */
            if((nick = xmlnode_get_data(x)) != NULL && j_strncasecmp(nick,"invite:",7) == 0)
            {
                nick += 7;
                if((jp->to = jid_new(jp->p, nick)) == NULL)
                {
                    jutil_error(jp->x,TERROR_BAD);
                }else{
                    xmlnode_put_attrib(jp->x,"to",jid_full(jp->to));
                    jp->from = jid_new(jp->p,jid_full(jid_user(from->localid)));
                    jid_set(jp->from,xmlnode_get_data(from->nick),JID_RESOURCE);
                    xmlnode_put_attrib(jp->x,"from",jid_full(jp->from));
                }
                deliver(dpacket_new(jp->x),NULL);
                return;
            }

	    /* Disallow subject change if user does not have high enough level */
	    if((!is_admin(room, from->realid) && room->subjectlock == 0) || is_visitor(room, from->realid) )
	    {
		jutil_error(jp->x,TERROR_FORBIDDEN);
		deliver(dpacket_new(jp->x),NULL);
		return;
	    }

	    /* Save room topic for new users */
            xmlnode_free(room->topic);
            room->topic = xmlnode_new_tag("topic");
            xmlnode_put_attrib(room->topic,"subject",xmlnode_get_data(x));
            xmlnode_insert_cdata(room->topic,xmlnode_get_data(from->nick),-1);
            xmlnode_insert_cdata(room->topic," has set the topic to: ",-1);
            xmlnode_insert_cdata(room->topic,xmlnode_get_data(x),-1);
        }

	/* Check if allowed to talk */
	if(room->moderated == 1 && !is_participant(room, from->realid))
	{
	    /* Attempting to talk in a moderated room without voice intrinsic */
            jutil_error(jp->x,TERROR_MUC_VOICE);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }

        /* ensure type="groupchat" */
        xmlnode_put_attrib(jp->x,"type","groupchat");

	/* Save copy of packet for history */
	node = xmlnode_dup(jp->x);

        /* broadcast */
        xmlnode_put_vattrib(jp->x,"cnu",(void*)from);
        xhash_walk(room->local,con_room_sendwalk,(void*)jp->x);

        /* log */
        con_room_log(room, xmlnode_get_data(from->nick), xmlnode_get_tag_data(jp->x, "body"));


	/* Save from address */
	id = jid_new(xmlnode_pool(node), jid_full(jid_user(from->localid)));
	jid_set(id, xmlnode_get_data(from->nick), JID_RESOURCE);
	xmlnode_put_attrib(node, "from", jid_full(id));

        /* store in history */
        jutil_delay(node, jid_full(room->id));
        if(room->master->history > 0)
        {
            if(++room->hlast == room->master->history)
                room->hlast = 0;
            xmlnode_free(room->history[room->hlast]);
            room->history[room->hlast] = node;
        }
            
	xmlnode_free(jp->x);
        return;
    }

    /* public iq requests */

    if(jpacket_subtype(jp) == JPACKET__SET)
    {
	if(NSCHECK(jp->iq, NS_MUC_ADMIN))
	{ 
	    log_debug(NAME, "[%s] IQ Set for admin function: >%s<", FZONE, xmlnode_get_name(jp->iq));

	    if(!is_admin(room, from->realid))
	    {
	        jutil_error(jp->x, TERROR_FORBIDDEN);
	        deliver(dpacket_new(jp->x), NULL);
	        return;
	    }

            if(j_strcmp(xmlnode_get_name(jp->iq), "query") == 0)
	    {
	        log_debug(NAME, "[%s] List set requested by admin...", FZONE);

		result = xmlnode_get_tag(jp->x, "query");

		if(NSCHECK(xmlnode_get_tag(result,"x"),NS_DATA))
		{
		    log_debug(NAME, "[%s] Received x:data", FZONE);
	            jutil_error(jp->x, TERROR_BAD);
		    deliver(dpacket_new(jp->x), NULL);
		    return;
		}
		else
		{
	            con_parse_item(from, jp);
		    return;
		}
            }
	}
	else if(NSCHECK(jp->iq, NS_MUC_OWNER))
	{
	    if(!is_owner(room, from->realid))
	    {   
		jutil_error(jp->x, TERROR_NOTALLOWED);
		deliver(dpacket_new(jp->x), NULL);
		return;
	    }

            if(j_strcmp(xmlnode_get_name(jp->iq),"destroy") == 0)
            {
		log_debug(NAME, "[%s] IQ Set for owner: destroy requested", FZONE);

		result = xmlnode_get_tag(jp->x, "destroy");

		xhash_walk(room->remote, con_room_leaveall, result);

	        con_room_zap(room);
	        jutil_iqresult(jp->x);
		deliver(dpacket_new(jp->x), NULL);
		return;
	    }
	    else if(j_strcmp(xmlnode_get_name(jp->iq),"query") == 0)
            {
	        log_debug(NAME, "[%s] IQ Set for owner: configuration set", FZONE);

		result = xmlnode_get_tag(jp->x, "query");

		if(NSCHECK(xmlnode_get_tag(result,"x"),NS_DATA))
		{
		    log_debug(NAME, "[%s] Received x:data", FZONE);
		    xdata_handler(room, from, jp);

		    jutil_iqresult(jp->x);
		    deliver(dpacket_new(jp->x), NULL);
		    return;
		}
		else
		{
	            con_parse_item(from, jp);
		    return;
		}
	    }
	    else
            {
	        jutil_error(jp->x, TERROR_BAD);
	        deliver(dpacket_new(jp->x), NULL);
	        return;
	    }
	}

        jutil_error(jp->x, TERROR_BAD);
        deliver(dpacket_new(jp->x), NULL);
        return;
    }

    if(jpacket_subtype(jp) == JPACKET__GET)
    {
	if(NSCHECK(jp->iq,NS_VERSION))
	{
	    jutil_iqresult(jp->x);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_VERSION);
	    jpacket_reset(jp);

	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"name"),NAME,-1);
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"version"),VERSION,-1);

	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}
	else if(NSCHECK(jp->iq, NS_BROWSE))
        {
            jutil_iqresult(jp->x);
            result = xmlnode_insert_tag(jp->x,"conference");
            xmlnode_put_attrib(result,"xmlns",NS_BROWSE);
            xmlnode_put_attrib(result,"name",room->name);

            if(room->public)
            {
                xmlnode_put_attrib(result,"type","public");
            }
	    else
	    {
                xmlnode_put_attrib(result,"type","private");
            }

            xmlnode_insert_cdata(xmlnode_insert_tag(result,"ns"),NS_MUC,-1);

            xhash_walk(room->local, con_room_browsewalk, (void*)result);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }
	else if(NSCHECK(jp->iq, NS_DISCO_INFO))
	{
	    log_debug(NAME, "[%s] room packet - Disco Request", FZONE);

	    jutil_iqresult(jp->x);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"), "xmlns", NS_DISCO_INFO);
	    jpacket_reset(jp);

	    result = xmlnode_insert_tag(jp->iq,"identity");
	    xmlnode_put_attrib(result, "category", "conference");
	    xmlnode_put_attrib(result, "type", "text");
	    xmlnode_put_attrib(result, "name", room->name);

	    xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "category", NS_GROUPCHAT);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "category", NS_MUC);

	    deliver(dpacket_new(jp->x), NULL);
	    return;
	}
	else if(NSCHECK(jp->iq, NS_MUC_ADMIN))
        {
	    log_debug(NAME, "[%s] IQ Get for admin function, %s", FZONE, xmlnode_get_name(jp->iq));

	    if(!is_admin(room, from->realid))
	    {
	        jutil_error(jp->x, TERROR_FORBIDDEN);
	        deliver(dpacket_new(jp->x), NULL);
	        return;
	    }

	    if(j_strcmp(xmlnode_get_name(jp->iq),"query") == 0)
            {
	        con_parse_item(from, jp);
	        return;
	    }
        }
	else if(NSCHECK(jp->iq, NS_MUC_OWNER))
        {
	    log_debug(NAME, "[%s] IQ Get for owner function, %s", FZONE, xmlnode_get_name(jp->iq));

	    if(!is_owner(room, from->realid))
	    {
	        jutil_error(jp->x, TERROR_FORBIDDEN);
	        deliver(dpacket_new(jp->x), NULL);
	        return;
	    }

	    if(j_strcmp(xmlnode_get_name(jp->iq),"query") == 0)
            {
	        con_parse_item(from, jp);
	        return;
	    }
        }
    }

    jutil_error(jp->x, TERROR_BAD);
    deliver(dpacket_new(jp->x), NULL);

}

cnr con_room_new(cni master, jid roomid, jid owner, char *name, char *secret, int private, int xdata, int persist)
{
    cnr room;
    pool p;
    cnu admin;
    char *key;

    p = pool_new();
    room = pmalloco(p, sizeof(_cnr));
    room->p = p;
    room->master = master;
    /* room jid */
    room->id = jid_new(p, jid_full(roomid));
    /* room name for browse */
    room->name = pstrdup(p, room->id->user);
    /* room password */
    room->secret = pstrdup(p, secret);
    room->private = private;
    room->history = pmalloco(p, sizeof(xmlnode) * master->history); /* make array of xmlnodes */
    /* user hashes */
    room->remote = xhash_new(master->maxhash);
    room->local = xhash_new(master->maxhash);
    room->roster = xhash_new(master->maxhash);
    /* Affiliated hashes */
    room->admin = xhash_new(master->maxhash);
    room->member = xhash_new(master->maxhash);
    room->outcast = xhash_new(master->maxhash);
    /* Role hashes */
    room->moderator = xhash_new(master->maxhash);
    room->participant = xhash_new(master->maxhash);
    /* Room messages */
    room->note_leave = xmlnode_get_tag_data(master->config,"notice/leave");
    room->note_join = xmlnode_get_tag_data(master->config,"notice/join");
    room->note_rename = xmlnode_get_tag_data(master->config,"notice/rename");
    /* Room Defaults */
    room->public = master->public;
    room->subjectlock = 0;
    room->maxusers = 30;
    room->persistant = persist;
    room->moderated = 0;
    room->defaulttype = 0;
    room->privmsg = 0;
    room->invitation = 0;
    room->invites = 0;
    room->legacy = 1;
    room->visible = 0;
    room->logfile = NULL;
    room->logformat = LOG_TEXT;
    room->description = NULL;


    if(owner != NULL)
    {
        admin = (void*)con_user_new(room, owner);
	add_roster(room, admin->realid);

	room->owner = jid_new(p, jid_full(owner));

	if(xdata > 0 )
		xdata_room_config(room,admin,1,NULL);

	log_debug(NAME, "[%s] Added new admin: %s to room %s", FZONE, jid_full(owner),jid_full(room->id));
    }

    key = j_strdup(jid_full(room->id));
    xhash_put(master->rooms, key, (void*)room);
    log_debug(NAME,"[%s] new room %s (%s) %s/%s/%d", FZONE, jid_full(room->id),jid_full(owner),name,secret,private);

    if(room->persistant == 1)
    {
	xdb_room_set(room);
    }

    return room;
}

void _con_room_send(xht h, const char *key, void *data, void *arg)
{
    cnu user = (cnu)data;
    xmlnode x = xmlnode_dup((xmlnode)arg);
    xmlnode_put_attrib(x,"to",jid_full(user->realid));
    deliver(dpacket_new(x),NULL);
}


void con_room_send(cnr room, xmlnode x)
{
    log_debug(NAME,"[%s] Sending packet from room %s: %s", FZONE, jid_full(room->id), xmlnode2str(x));

    /* log */
    con_room_log(room, NULL, xmlnode_get_tag_data(x, "body"));

    xmlnode_put_attrib(x,"from",jid_full(room->id));
    xhash_walk(room->local,_con_room_send,(void*)x);
    xmlnode_free(x);
}


void con_room_zap(cnr room)
{
    log_debug(NAME, "[%s] zapping room %s", FZONE, jid_full(room->id));

    /* Clear old hashes */
    xhash_free(room->remote);
    xhash_free(room->local);
    xhash_free(room->roster);
    xhash_free(room->admin);
    xhash_free(room->member);
    xhash_free(room->outcast);
    xhash_free(room->moderator);
    xhash_free(room->participant);
    if(room->logfile)
	fclose(room->logfile);

    xhash_zap(room->master->rooms,jid_full(room->id));
    xmlnode_free(room->topic);
    pool_free(room->p);
}

